function flow = ComputeSimilarity(img1, img2)

%%% This function compute similarity transform and output the similarity
%%% flow between the two images. The flow would be from img2 to img1.

%%% Declaring global variables ---------
global patch_w;
global DSMethod;
global maxRansacFeatures;
global maxImgSize;
%%% ------------------------------------

[h, w, ~] = size(img1);
imgSize = h * w;
[hTest, wTest, ~] = size(img2);

assert(h == hTest && w == wTest, 'The size of the input frames should be the same');

%%% This part tries to downsample the input image if it is big. The
%%% similarity can be computed on the downsampled image and the
%%% transformation can be scaled and used on the original image. Also we
%%% don't use flows at every pixel to compute the similarity transform. The
%%% skipFac defines how many pixels should be skipped. The whole thing is
%%% just for speed up
if (imgSize >= maxImgSize)
    resizeFac = sqrt(maxImgSize / imgSize);
    img1 = imresize(img1, resizeFac, DSMethod);
    img2 = imresize(img2, resizeFac, DSMethod);

    sh = ceil(h * resizeFac);
    sw = ceil(w * resizeFac);
    
    sImgSize = sh * sw;
else
    resizeFac = 1;
    sh = h;
    sw = w;
    
    sImgSize = sh * sw;
end

skipFac = floor(sqrt(sImgSize / maxRansacFeatures));
%%% ------------------------------------------------------------------

[flowBeginX, flowBeginY, flowEndX, flowEndY] = ComputeCorrespondences(img2, img1);

%%% Since PatchMatch correspondence is missing right and bottom boundaries by
%%% the size of the patch, we shouldn't take thouse regions into account.
%%% Therefore we subtract patch_w-1 from the end of the matrix.
flowBeginX = Vec(flowBeginX(1 : skipFac : end-patch_w+1, 1 : skipFac : end-patch_w+1));
flowBeginY = Vec(flowBeginY(1 : skipFac : end-patch_w+1, 1 : skipFac : end-patch_w+1));
flowEndX = Vec(flowEndX(1 : skipFac : end-patch_w+1, 1 : skipFac : end-patch_w+1));
flowEndY = Vec(flowEndY(1 : skipFac : end-patch_w+1, 1 : skipFac : end-patch_w+1));


%%% For speed up in calculation of the similarity transform, we first
%%% compute the index of matches which can be fitted to a loos translation
%%% model. Then we compute the similarity model only on these indecies with
%%% a tight sigma parameter. 
sigmaTranslation = 2.5 * sqrt(sh*sw)/500;
[~, inds] = FitModel(flowBeginX, flowBeginY, flowEndX, flowEndY, sigmaTranslation, 'Translation');

sigmaSimilarity = 0.5 * sqrt(sh*sw)/500;
[H, ~] = FitModel(flowBeginX(inds), flowBeginY(inds), flowEndX(inds), flowEndY(inds), sigmaSimilarity, 'similarity');

%%% Since the output model has only two rows, we pad it by [0, 0, 1]. We
%%% also need to scale the model to the original size since we calculated
%%% it on the downsampled image.
H = reshape(H, 2, 3);
H = [H; [0, 0, 1]];
H = [1/resizeFac 0 0;0 1/resizeFac 0;0 0 1;] * H / [1/resizeFac 0 0;0 1/resizeFac 0;0 0 1;];
%%% --------------------------------------------------------------------

flow = MatrixToFlow(H, [h, w]);

