function [Theta, k] = estimate_affine(X, s, transform_type)

% [Theta k] = estimate_affine(X, s)
%
% DESC:
% estimate the parameters of an affine trasformation via least squares.
%
% X2 = A * X1 + b
%
% where Theta = [a11; a21; a12; a22; b1; b2];

%
% VERSION:
% 1.0.0
%
% INPUT:
% X                 = 2D point correspondences
% s                 = indices of the points used to estimate the parameter
%                     vector. If empty all the points are used
%
% OUTPUT:
% Theta             = estimated parameter vector 
% k                 = dimension of the minimal subset


% AUTHOR:
% Marco Zuliani, email: marco.zuliani@gmail.com
% Copyright (C) 2009 by Marco Zuliani 
% MODIFIED
% By Nima Khademi Kalantari
% 
% LICENSE:
% This toolbox is distributed under the terms of the GNU GPL.
% Please refer to the files COPYING.txt for more information.


% HISTORY:
% 1.0.0             = 08/27/08 - initial version

% cardinality of the MSS
switch lower(transform_type)
	case {'similarity'}
		k = 3;
	case 'translation'
		k = 1;
	otherwise 
		error('Unknown transofrmation type %s.', params.transform_type)
end

if (nargin == 0) || isempty(X)
    Theta = [];
    return;
end;

if (nargin > 2) && ~isempty(s)
    X = X(:, s);
end;

if (nargin == 2 || (nargin == 3 && isempty(transform_type)))
	transform_type = 'similarity';
end

    
% check if we have enough points
N = size(X, 2);
if (N < k)
    error('estimate_RST:inputError', ...
        'At least 2 point correspondences are required');
end;

X1 = X(1:2, :);
X2 = X(3:4, :);

switch lower(transform_type)
	case {'similarity'}
        tr = cp2tform(X1', X2', 'nonreflective similarity');
        if (isfield(tr.tdata, 'T'))
            H = tr.tdata.T';
			H = H(1:2, :);
        else
            H = [];
        end
	case 'translation'
		H = mean(X2-X1, 2);
		H = [1 0 H(1); 0 1 H(2)];
end

Theta = H(:);

return;